/*
 * This program is based on two examples provided by Greg Kroah-Hartman:
 *
 *   USB LED driver - 1.1
 *   Copyright (C) 2004 Greg Kroah-Hartman (greg@kroah.com)
 *
 *   USB Skeleton driver - 2.0
 *   Copyright (C) 2001-2004 Greg Kroah-Hartman (greg@kroah.com)
 *
 *        This program is free software; you can redistribute it and/or
 *        modify it under the terms of the GNU General Public License as
 *        published by the Free Software Foundation, version 2.
 *
 * This document by Patrick Mochel was also very helpful:
 * http://www.mjmwired.net/kernel/Documentation/filesystems/sysfs.txt
 *
 */

#include <linux/kernel.h>
#include <linux/ctype.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/usb.h>

#define DRIVER_AUTHOR "J. Mike Rollins <rollins@CamoTruck.Net>"
#define DRIVER_DESC "USB driver for data meter"

#define VENDOR_ID	0x04d8
#define PRODUCT_ID	0x000c

#define USBGEN_EP_SIZE  64
#define KHz  *1000

#include "user_common.h"

#define DEBUG_ON 0

#define MAX_BUFFER_STRING_LEN 24


/* table of devices that work with this driver */
static struct usb_device_id id_table [] = {
	{ USB_DEVICE(VENDOR_ID, PRODUCT_ID) },
	{ },
};
MODULE_DEVICE_TABLE (usb, id_table);

/* Structure to hold all of our device specific stuff */
struct usb_taz_data_meter {
        struct usb_device *     udev;                   /* the usb device for this device */
        struct usb_interface *  interface;              /* the interface for this device */
        size_t                  bulk_in_size;           /* the size of the receive buffer */
        __u8                    bulk_in_endpointAddr;   /* the address of the bulk in endpoint */
        __u8                    bulk_out_endpointAddr;  /* the address of the bulk out endpoint */
        unsigned long           motor_delay;
        int                     motor_m1;
        char                    out_string[USBGEN_EP_SIZE];         
        char                    in_string[USBGEN_EP_SIZE];         
};


static ssize_t usb_send_data(struct device*, struct device_attribute*, const char *, size_t);
static ssize_t usb_get_data(struct device*, struct device_attribute*, char *, size_t);
static int my_strtoint(const char *s, int len);



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
static ssize_t data_meter_show_SIZE(struct device *dev, struct device_attribute *attr, char *buf)
  {
    int bytes;
    CMD_TYPE command;
    DATA_PACKET buffer;
    UINT32 capacity;
    UINT32 sector_size;

    #if DEBUG_ON
      dev_info(dev, "Sending SIZE request to USB device\n");
    #endif

    pack_C( &buffer, CMD_REQ_SIZE);

    bytes =  usb_send_data(dev, attr, (unsigned char*) &buffer, USBGEN_EP_SIZE);
    if (bytes != USBGEN_EP_SIZE) return -EFAULT;

    memset(&buffer,0,USBGEN_EP_SIZE);

    #if DEBUG_ON
      dev_info(dev, "Reading answer from USB device\n");
    #endif

    bytes =  usb_get_data(dev, attr, (unsigned char*) &buffer, DATA_PACKET_SIZE);

    if (bytes != DATA_PACKET_SIZE) return -EFAULT;

    command = buffer.b[0];

    #if DEBUG_ON
      dev_info(dev, "%.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x\n",
      buffer.b[0], buffer.b[1], buffer.b[2], buffer.b[3], buffer.b[4], buffer.b[5], 
      buffer.b[6], buffer.b[7], buffer.b[8], buffer.b[9]);
    #endif

    //
    // read response
    if (command == CMD_ANS_DATA) 
      {
        // NOTE: buf is PAGE_SIZE (about 4K)
        unpack_C44(&buffer,&command, &capacity, &sector_size);

        return sprintf(buf, "Capacity = %ld\nSectorSize = %ld\n", capacity, sector_size);
      }
    else
      {
        dev_info(dev, "Bad response from USB device\n");
        return -EFAULT;
      }

  }

static ssize_t data_meter_store_SIZE(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
  {
    //struct usb_interface *intf = to_usb_interface(dev);
    //struct usb_taz_data_meter *taz_data_meter = usb_get_intfdata(intf);
    return count;
  }

static struct device_attribute dev_attr_SIZE = {
       .attr    = {
                .name = "sd_size",
                .mode = S_IWUSR | S_IRUGO,
        },
        .show = data_meter_show_SIZE,
        .store = data_meter_store_SIZE
};



/*******************/


static ssize_t data_meter_show_PING(struct device *dev, struct device_attribute *attr, char *buf)
  {
    int bytes;
    CMD_TYPE command;
    DATA_PACKET buffer;

    #if DEBUG_ON
      dev_info(dev, "Sending PING request to USB device\n");
    #endif

    pack_C( &buffer, CMD_REQ_PING);

    bytes =  usb_send_data(dev, attr, (unsigned char*) &buffer, USBGEN_EP_SIZE);
    if (bytes != USBGEN_EP_SIZE) return -EFAULT;

    memset(&buffer,0,USBGEN_EP_SIZE);

    #if DEBUG_ON
      dev_info(dev, "Reading answer from USB device\n");
    #endif

    bytes =  usb_get_data(dev, attr, (unsigned char*) &buffer, DATA_PACKET_SIZE);

    if (bytes != DATA_PACKET_SIZE) return -EFAULT;

    command = buffer.b[0];

    // read response
    if (command != CMD_ANS_PONG) 
      {
        dev_info(dev, "Bad response from USB device\n");
        return -EFAULT;
      }

    // NOTE: buf is PAGE_SIZE (about 4K)
    return 
      sprintf(buf, "%s\n", "pong");
  }

static ssize_t data_meter_store_PING(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
  {
    //struct usb_interface *intf = to_usb_interface(dev);
    //struct usb_taz_data_meter *taz_data_meter = usb_get_intfdata(intf);
    return count;
  }

static struct device_attribute dev_attr_PING = {
       .attr    = {
                .name = "ping",
                .mode = S_IWUSR | S_IRUGO,
        },
        .show = data_meter_show_PING,
        .store = data_meter_store_PING
};

/**************************************************************************/


static ssize_t data_meter_show_SDTEST(struct device *dev, struct device_attribute *attr, char *buf)
  {
    int bytes;
    UINT16 delay;
    UINT16 value;
    CMD_TYPE command;
    DATA_PACKET buffer;

    pack_C( &buffer, CMD_REQ_SDTEST_GET);

    #if DEBUG_ON
      dev_info(dev, "Sending SDTEST_GET request to USB device\n");
    #endif

    bytes =  usb_send_data(dev, attr, (unsigned char*)&buffer, DATA_PACKET_SIZE);
    if (bytes != DATA_PACKET_SIZE) return -EFAULT;

    memset(&buffer,0,USBGEN_EP_SIZE);

    #if DEBUG_ON
      dev_info(dev, "Reading SDTEST_GET answer from USB device\n");
    #endif

    bytes =  usb_get_data(dev, attr, (unsigned char*) &buffer, DATA_PACKET_SIZE);

    if (bytes != DATA_PACKET_SIZE) return -EFAULT;

    command = buffer.b[0];

    #if DEBUG_ON
      dev_info(dev, "%.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x\n",
      buffer.b[0], buffer.b[1], buffer.b[2], buffer.b[3], buffer.b[4], 
      buffer.b[5], buffer.b[6], buffer.b[7], buffer.b[8], buffer.b[9]);
    #endif

    // read response
    if (command == CMD_ANS_DATA) 
      {
        unpack_C22(&buffer, &command, &delay, &value);
        // NOTE: buf is PAGE_SIZE (about 4K)
        return sprintf(buf, "%d, %d\n", delay, value);
      }
    else if (command == CMD_ERR_EOF) 
      {
        dev_info(dev, "Received an CMD_ERR_EOF from USB device\n");
        return -EFAULT;
      }
    else 
      {
        dev_info(dev, "Bad response from USB device\n");
        return -EFAULT;
      }
  }


static ssize_t data_meter_store_SDTEST(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
  {
    // unsigned char buffer[USBGEN_EP_SIZE];
    int bytes;
    char* ptr;
    int len;
    unsigned int delay = 0;
    unsigned int value = 0;
    CMD_TYPE command;
    DATA_PACKET buffer;

    delay = my_strtoint(buf,count);

    for(ptr=(char*)buf, len=count; (len>0) && (*ptr != ',') ; len--,ptr++);
    len--;
    ptr++;
    value = my_strtoint(ptr,len);
    
    #if DEBUG_ON
      dev_info(dev, "Read '%d' and '%d' from file\n", delay, value);
    #endif

    pack_C22(&buffer, CMD_REQ_SDTEST_PUT, delay, value);

    #if DEBUG_ON
      dev_info(dev, "Sending SD card value to USB device\n");
    #endif

    bytes =  usb_send_data(dev, attr, (unsigned char*)&buffer, DATA_PACKET_SIZE);
    if (bytes != DATA_PACKET_SIZE) return -EFAULT;

    memset(&buffer,0,USBGEN_EP_SIZE);

    #if DEBUG_ON
      dev_info(dev, "Reading answer from USB device\n");
    #endif

    bytes =  usb_get_data(dev, attr, (unsigned char*) &buffer, DATA_PACKET_SIZE);

    if (bytes != DATA_PACKET_SIZE) return -EFAULT;

    command = buffer.b[0];

    // read response
    if (command == CMD_OK) 
      {
        // NOTE: buf is PAGE_SIZE (about 4K)
        return count;
      }
    else if (command == CMD_ERR) 
      {
        dev_info(dev, "Received an CMD_ERR from USB device\n");
        return -EFAULT;
      }
    else 
      {
        dev_info(dev, "Bad response from USB device\n");
        return -EFAULT;
      }
    return count;
  }


static struct device_attribute dev_attr_SDTEST = {
       .attr    = {
                .name = "sd_test",
                .mode = S_IWUSR | S_IRUGO,
        },
        .show = data_meter_show_SDTEST,
        .store = data_meter_store_SDTEST
};

/***********************************************************/

static ssize_t data_meter_show_PTEST(struct device *dev, struct device_attribute *attr, char *buf)
  {
    int bytes;
    UINT16 value;
    CMD_TYPE command;
    DATA_PACKET buffer;

    pack_C( &buffer, CMD_GET_PTEST);

    #if DEBUG_ON
      dev_info(dev, "Sending CMD_GET_PTEST request to USB device\n");
    #endif

    bytes =  usb_send_data(dev, attr, (unsigned char*)&buffer, DATA_PACKET_SIZE);
    if (bytes != DATA_PACKET_SIZE) return -EFAULT;

    memset(&buffer,0,USBGEN_EP_SIZE);

    #if DEBUG_ON
      dev_info(dev, "Reading answer from USB device\n");
    #endif

    bytes =  usb_get_data(dev, attr, (unsigned char*) &buffer, DATA_PACKET_SIZE);

    if (bytes != DATA_PACKET_SIZE) return -EFAULT;

    command = buffer.b[0];

    #if DEBUG_ON
      dev_info(dev, "%.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x\n",
      buffer.b[0], buffer.b[1], buffer.b[2], buffer.b[3], buffer.b[4], 
      buffer.b[5], buffer.b[6], buffer.b[7], buffer.b[8], buffer.b[9]);
    #endif

    // read response
    if (command == CMD_VAL_PTEST) 
      {
        unpack_C2(&buffer, &command, &value);
        // NOTE: buf is PAGE_SIZE (about 4K)
        return sprintf(buf, "%d\n", value);
      }
    else if (command == CMD_ERR_EOF) 
      {
        dev_info(dev, "Received an CMD_ERR_EOF from USB device\n");
        return -EFAULT;
      }
    else 
      {
        dev_info(dev, "Bad response from USB device\n");
        return -EFAULT;
      }
  }

static ssize_t data_meter_store_PTEST(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
{
    // unsigned char buffer[USBGEN_EP_SIZE];
    int bytes;
    unsigned int value = 0;
    CMD_TYPE command;
    DATA_PACKET buffer;

    value = my_strtoint(buf,count);

    #if DEBUG_ON
      dev_info(dev, "Read '%d' from file\n", value);
    #endif

    pack_C2(&buffer, CMD_SET_PTEST, value);

    #if DEBUG_ON
      dev_info(dev, "Sending ptest value to USB device\n");
    #endif

    bytes =  usb_send_data(dev, attr, (unsigned char*)&buffer, DATA_PACKET_SIZE);
    if (bytes != DATA_PACKET_SIZE) return -EFAULT;

    memset(&buffer,0,USBGEN_EP_SIZE);

    #if DEBUG_ON
      dev_info(dev, "Reading answer from USB device\n");
    #endif

    bytes =  usb_get_data(dev, attr, (unsigned char*) &buffer, DATA_PACKET_SIZE);

    if (bytes != DATA_PACKET_SIZE) return -EFAULT;

    command = buffer.b[0];

    // read response
    if (command == CMD_OK) 
      {
        // NOTE: buf is PAGE_SIZE (about 4K)
        return count;
      }
    else if (command == CMD_ERR) 
      {
        dev_info(dev, "Received an CMD_ERR from USB device\n");
        return -EFAULT;
      }
    else 
      {
        dev_info(dev, "Bad response from USB device\n");
        return -EFAULT;
      }
    return count;
  }

static struct device_attribute dev_attr_PTEST = {
       .attr    = {
                .name = "protocol_test",
                .mode = S_IWUSR | S_IRUGO,
        },
        .show = data_meter_show_PTEST,
        .store = data_meter_store_PTEST
};


/**************************************************************************/

static ssize_t data_meter_show_SEC_COUNT(struct device *dev, struct device_attribute *attr, char *buf)
{
    int bytes;
    DATA_PACKET buffer;
    UINT32 sector_count;
    CMD_TYPE command;

    // memset(&packet,0,DATA_PACKET_SIZE);
    // packet.command = CMD_REQ_SEC_COUNT;

    pack_C(&buffer,CMD_REQ_SEC_COUNT);

    #if DEBUG_ON
      dev_info(dev, "Sending sector_count request to USB device\n");
    #endif

    bytes =  usb_send_data(dev, attr, (unsigned char*)&buffer, DATA_PACKET_SIZE);
    if (bytes != DATA_PACKET_SIZE) return -EFAULT;

    memset(&buffer,0,USBGEN_EP_SIZE);

    #if DEBUG_ON
      dev_info(dev, "Reading sector_count answer from USB device\n");
    #endif

    bytes =  usb_get_data(dev, attr, (unsigned char*) &buffer, DATA_PACKET_SIZE);

    #if DEBUG_ON
      dev_info(dev, "%.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x\n",
      buffer.b[0], buffer.b[1], buffer.b[2], buffer.b[3], buffer.b[4], buffer.b[5], 
      buffer.b[6], buffer.b[7], buffer.b[8], buffer.b[9]);
    #endif


    if (bytes != DATA_PACKET_SIZE) return -EFAULT;

    // memcpy(&packet, buffer, DATA_PACKET_SIZE );
    command = buffer.b[0];

    // read response
    if (command == CMD_ANS_DATA) 
      {
        unpack_C4(&buffer,&command,&sector_count);
        // NOTE: buf is PAGE_SIZE (about 4K)
        return sprintf(buf, "%ld\n", sector_count);
      }
    else 
      {
        dev_info(dev, "Bad response from USB device (%d)\n",command);
        return -EFAULT;
      }
  }

static ssize_t data_meter_store_SEC_COUNT(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
  {
    //struct usb_interface *intf = to_usb_interface(dev);
    //struct usb_taz_data_meter *taz_data_meter = usb_get_intfdata(intf);
    return count;
  }

static struct device_attribute dev_attr_SEC_COUNT = {
       .attr    = {
                .name = "sd_sector_count",
                .mode = S_IWUSR | S_IRUGO,
        },
        .show = data_meter_show_SEC_COUNT,
        .store = data_meter_store_SEC_COUNT
};

/* * * */

static ssize_t data_meter_show_BULK_DATA(struct device *dev, struct device_attribute *attr, char *buf)
  {
    int bytes;
    int i;
    char str[64];

    UINT16 delay;
    UINT16 value;
    CMD_TYPE command;
    DATA_PACKET buffer;

    pack_C( &buffer, CMD_REQ_BULK);

    #if DEBUG_ON
      dev_info(dev, "Sending CMD_REQ_BULK request to USB device\n");
    #endif

    bytes =  usb_send_data(dev, attr, (unsigned char*)&buffer, DATA_PACKET_SIZE);
    if (bytes != DATA_PACKET_SIZE) return -EFAULT;

    memset(&buffer,0,USBGEN_EP_SIZE);

    #if DEBUG_ON
      dev_info(dev, "Reading CMD_REQ_BULK answer from USB device\n");
    #endif

    bytes =  usb_get_data(dev, attr, (unsigned char*) &buffer, DATA_PACKET_SIZE);

    if (bytes != DATA_PACKET_SIZE) return -EFAULT;

    command = buffer.b[0];
    
    if (command == CMD_ANS_BULK)
      {
        buf[0] = '\0';

        #if DEBUG_ON
          dev_info(dev, "%.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x\n",
          buffer.b[0], buffer.b[1], buffer.b[2], buffer.b[3], buffer.b[4], buffer.b[5], 
          buffer.b[6], buffer.b[7], buffer.b[8], buffer.b[9]);
        #endif

        unpack_bulk_C(&buffer, &command);
        for (i=0; i<15; i++)
          {
            // NOTE: buf is PAGE_SIZE (about 4K)
            unpack_bulk_n22(&buffer,i,&delay,&value);
            sprintf(str,"%d,%d\n",delay,value);
            strcat(buf,str);

          }
        return strlen(buf);
      }
    else if (command == CMD_ERR_EOF)
      {
        dev_info(dev, "Received an CMD_ERR_EOF from USB device\n");
        return -EFAULT;
      }
    else 
      {
        dev_info(dev, "Bad response from USB device\n");
        return -EFAULT;
      }
  }



// next revision, allow integer write to file to indicate move to a particular sector
static ssize_t data_meter_store_BULK_DATA(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
  {
    int bytes;
    CMD_TYPE command;
    DATA_PACKET buffer;


    if (buf[0] == 'E')
      {
        // send erase command
        pack_C(&buffer, CMD_REQ_ERASE);
        #if DEBUG_ON
          dev_info(dev, "Sending ERASE command to USB device\n");
        #endif
      }
    else
      {
        // send erase command
        #if DEBUG_ON
          dev_info(dev, "Sending SOF command to USB device\n");
        #endif

        pack_C(&buffer, CMD_REQ_SOF);
      }

    bytes =  usb_send_data(dev, attr, (unsigned char*)&buffer, DATA_PACKET_SIZE);
    if (bytes != DATA_PACKET_SIZE) return -EFAULT;

    memset(&buffer,0,USBGEN_EP_SIZE);

    #if DEBUG_ON
      dev_info(dev, "Reading answer from USB device\n");
    #endif

    bytes =  usb_get_data(dev, attr, (unsigned char*) &buffer, DATA_PACKET_SIZE);

    if (bytes != DATA_PACKET_SIZE) return -EFAULT;

    command = buffer.b[0];

    // read response
    if (command == CMD_OK) 
      {
        // NOTE: buf is PAGE_SIZE (about 4K)
        return count;
      }
    else if (command == CMD_ERR) 
      {
        dev_info(dev, "Received an CMD_ERR from USB device\n");
        return -EFAULT;
      }
    else 
      {
        dev_info(dev, "Bad response from USB device\n");
        return -EFAULT;
      }
    return count;
  }


static struct device_attribute dev_attr_BULK_DATA = {
       .attr    = {
                .name = "sd_data",
                .mode = S_IWUSR | S_IRUGO,
        },
        .show = data_meter_show_BULK_DATA,
        .store = data_meter_store_BULK_DATA
};




/** Utility Functions **/

static int my_strtoint(const char *s, int len)
  {
    int i;
    int value = 0;
    int sign = 0;
    char c;

    for (i=0; i<len; i++)
      {
        c = s[i];
        if (isdigit(c))
          {
            value = value*10 + (c-'0');
            continue;
          }

        if (value == 0)
          {
            if (isspace(c))
              continue;
          }

        if (sign == 0)
          {
            if (c == '+')
              {
                sign = 1;
                continue;
              }
            else if (c == '-')
              {
                sign = -1;
                continue;
              }
          }    
        break;
      }
    sign=(sign<0)?sign:(+1);
    return sign*value;
  }

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *  USB Communications Utility
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static ssize_t usb_send_data(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
  {
    struct usb_interface *intf = to_usb_interface(dev);
    struct usb_taz_data_meter *taz_data_meter = usb_get_intfdata(intf);

    int i,bytes,retval = 0;
    char *buffer;
    
    buffer = kzalloc(USBGEN_EP_SIZE, GFP_KERNEL);
    if (! buffer)
      {
        dev_info(dev, "Could not allocate bulk_in_buffer\n");
        goto error;
      }

    if (count > USBGEN_EP_SIZE)
      {
        dev_info(dev, "string data too long. count=%d\n",count);
        goto error;
      }

    /* copy the parameter into our new buffer */
    for (i=0; i<count; i++)
        buffer[i] = buf[i];

    /* do a blocking bulk write to send data to the device */
    retval = usb_bulk_msg(taz_data_meter->udev,
                          usb_sndbulkpipe(taz_data_meter->udev, taz_data_meter->bulk_out_endpointAddr),
                          buffer,
                          count,  
                          &bytes, 
                          5 KHz);

    if (retval)
      {
        dev_info(dev, "usb_bulk_msg write returned an error: %d\n",retval);
        retval = -EFAULT;
        goto error;
      }

    if ( !(bytes == count) )
      {
        dev_info(dev, "usb_bulk_msg write byte count mismatch (%d != %d)\n",bytes,count);
        retval = -EFAULT;
        goto error;
      }

    retval = bytes;

error:

    kfree(buffer);
    return retval;
  }

static ssize_t usb_get_data(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
  {
    struct usb_interface *intf = to_usb_interface(dev);
    struct usb_taz_data_meter *taz_data_meter = usb_get_intfdata(intf);

    int i;
    int bytes = 0;
    int retval = 0;
    char *buffer;
    
    buffer = kzalloc(USBGEN_EP_SIZE, GFP_KERNEL);
    if (! buffer)
      {
        dev_info(dev, "Could not allocate bulk_in_buffer\n");
        goto error;
      }

    if (count > USBGEN_EP_SIZE)
      {
        dev_info(dev, "string data too long. count=%d\n",count);
        goto error;
      }

    /* do a blocking bulk read to get data from the device */
    retval = usb_bulk_msg(taz_data_meter->udev,
                          usb_rcvbulkpipe(taz_data_meter->udev, taz_data_meter->bulk_in_endpointAddr),
                          buffer,
                          count,  
                          &bytes, 
                          5 KHz);

    if (retval)
      {
        dev_info(dev, "usb_bulk_msg read returned an error: %d\n",retval);
        retval = -EFAULT;
        goto error;
      }

    if ( !(bytes == count) )
      {
        dev_info(dev, "usb_bulk_msg read byte count mismatch (%d != %d)\n",bytes,count);
        retval = -EFAULT;
        goto error;
      }

    for (i=0; i<count; i++)
        buf[i] = buffer[i];

    retval = count;

error:

    kfree(buffer);
    return retval;
  }




static int taz_data_meter_probe(struct usb_interface *interface, const struct usb_device_id *id)
  {
    struct usb_device *udev = interface_to_usbdev(interface);
    struct usb_taz_data_meter *dev = NULL;
    int retval = -ENOMEM;

    /* usb setup */
    struct usb_host_interface *iface_desc;
    struct usb_endpoint_descriptor *endpoint;
    size_t buffer_size;
    int i;

    dev = kzalloc(sizeof(struct usb_taz_data_meter), GFP_KERNEL);
    if (dev == NULL) 
      {
        dev_err(&interface->dev, "Out of memory\n");
        goto error;
      }

    dev->udev = usb_get_dev(udev);

    /* sysfs setup */
    //
    // Linux version 2.6.29.4-167.fc11.i586:
    //   - warning: ignoring return value of ‘device_create_file’, declared with attribute warn_unused_result
    //
    device_create_file(&interface->dev, &dev_attr_PING);
    device_create_file(&interface->dev, &dev_attr_SDTEST);
    device_create_file(&interface->dev, &dev_attr_SEC_COUNT);
    device_create_file(&interface->dev, &dev_attr_BULK_DATA);
    device_create_file(&interface->dev, &dev_attr_SIZE);
    device_create_file(&interface->dev, &dev_attr_PTEST);


    /* default values */
    //dev->motor_delay = MOTOR_DELAY;

    /* usb setup */
    dev->interface = interface;

    /* set up the endpoint information */
    /* use only the first bulk-in and bulk-out endpoints */
    iface_desc = interface->cur_altsetting;
    for (i = 0; i < iface_desc->desc.bNumEndpoints; ++i) 
      {
        endpoint = &iface_desc->endpoint[i].desc;

        /* check for endpoints that are IN and BULK */
        if (!dev->bulk_in_endpointAddr &&
            ((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK)
                                    == USB_DIR_IN) &&
            ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
                                    == USB_ENDPOINT_XFER_BULK)) 
          {
            /* we found a bulk in endpoint */
            buffer_size = le16_to_cpu(endpoint->wMaxPacketSize);
            dev->bulk_in_size = buffer_size;
            dev->bulk_in_endpointAddr = endpoint->bEndpointAddress;
          }


        /* check for endpoints that are OUT and BULK */
        if (!dev->bulk_out_endpointAddr &&
            ((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK)
                                    == USB_DIR_OUT) &&
            ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
                                    == USB_ENDPOINT_XFER_BULK)) 
          {
            /* we found a bulk out endpoint */
            dev->bulk_out_endpointAddr = endpoint->bEndpointAddress;
          }
      } // for 

      if (!(dev->bulk_in_endpointAddr && dev->bulk_out_endpointAddr)) 
        {
          dev_info(&interface->dev, "Could not find both bulk-in and bulk-out endpoints\n");
          goto error;
        }

    /* save our data pointer in this interface device */
    usb_set_intfdata (interface, dev);

    /* let the user know what node this device is now attached to */
    dev_info(&interface->dev, 
        "USB taz_data_meter device now attached: EP_IN %d; EP_OUT %d; SIZE %d\n",
        dev->bulk_in_endpointAddr, 
        dev->bulk_out_endpointAddr, 
        dev->bulk_in_size);

    return 0;

error:
    kfree(dev);
    return retval;
}


static void taz_data_meter_disconnect(struct usb_interface *interface)
  {
    struct usb_taz_data_meter *dev;

    dev = usb_get_intfdata (interface);
    usb_set_intfdata (interface, NULL);
    
    device_remove_file(&interface->dev, &dev_attr_PING);
    device_remove_file(&interface->dev, &dev_attr_SDTEST);
    device_remove_file(&interface->dev, &dev_attr_SEC_COUNT);
    device_remove_file(&interface->dev, &dev_attr_BULK_DATA);
    device_remove_file(&interface->dev, &dev_attr_SIZE);
    device_remove_file(&interface->dev, &dev_attr_PTEST);

    usb_put_dev(dev->udev);
    
    kfree(dev);
    
    dev_info(&interface->dev, "USB taz_data_meter now disconnected\n");
  }

static struct usb_driver taz_data_meter_driver = {
	.name =		"taz_data_meter",
	.probe =	taz_data_meter_probe,
	.disconnect =	taz_data_meter_disconnect,
	.id_table =	id_table,
};

static int __init usb_taz_data_meter_init(void)
  {
    int retval = 0;

    retval = usb_register(&taz_data_meter_driver);
    if (retval)
        err("usb_register failed. Error number %d", retval);
    return retval;
  }

static void __exit usb_taz_data_meter_exit(void)
  {
    usb_deregister(&taz_data_meter_driver);
  }

module_init (usb_taz_data_meter_init);
module_exit (usb_taz_data_meter_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL v2");

