
/** INCLUDES *******************************************************/

#include <p18f4550.h>
#include <timers.h>
#include <pwm.h>
#include "protocol.h"

// each time Timer0 overflows, the protocol_value will be assigned a value.
volatile unsigned int protocol_value = 0;


void SetDC(unsigned int dutycycle)
{
  union PWMDC DCycle;
  
  // The PWM duty cycle is specified by writing to the
  // CCPR1L register and to the CCP1CON<5:4> bits. Up
  // to 10-bit resolution is available. The CCPR1L contains
  // the eight MSbs and the CCP1CON<5:4> contains the
  // two LSbs. This 10-bit value is represented by
  // CCPR1L:CCP1CON<5:4>.

  // Save the dutycycle value in the union
  DCycle.lpwm = dutycycle << 6;

  // Write the high byte into CCPR1L
  CCPR1L = DCycle.bpwm[1];

  // Write the low byte into CCP1CON5:4
  CCP1CON = (CCP1CON & 0xCF) | ((DCycle.bpwm[0] >> 2) & 0x30);
}



void protocol_init(void)
  {
    // Timer1 counts the pulses from the external circuit.
    OpenTimer1( 
        TIMER_INT_ON & 
        T1_16BIT_RW &      // max count of 65535    
        T1_SOURCE_EXT &  
        T1_PS_1_1 &        // options are 1,2,4,8
        T1_OSC1EN_OFF &
        T1_SYNC_EXT_OFF 
      );
 
    /***** Configure PWM ******/
    
    // 1. Configure the PWM pins, P1A and P1B (and
    //    P1C and P1D, if used), as inputs by setting the
    //    corresponding TRIS bits.
  
    TRISCbits.TRISC2 = 1; // P1A
    TRISDbits.TRISD5 = 1; // P1B
       
    // 2. Set the PWM period by loading the PR2 register.
    //    TOSC = 1/Clock Frequency
    //    PWM Period = [(PR2) + 1]  4  TOSC  (TMR2 Prescale Value)
    
    PR2 = PCMD_IDLE;    // 0 to 255
    
    // 4. Configure the ECCP module for the desired
    //    PWM mode and configuration by loading the
    //    CCP1CON register with the appropriate values:
    //    - Select one of the available output
    //      configurations and direction with the
    //      P1M1:P1M0 bits.
    //    - Select the polarities of the PWM output
    //      signals with the CCP1M3:CCP1M0 bits.
    
    // CCP1CON 7:6  (Half-Bridge) = 10
    
    CCP1CON = 0b10001100; // half bridge/active high
    //CCP1CON = 0b10001111; // half bridge/active low
    
    // 5. Set the PWM duty cycle by loading the CCPR1L
    //    register and CCP1CON<5:4> bits.
    
    SetDC(2*(PCMD_IDLE+1)); // 0 to 1023
    
    // 6. For Half-Bridge Output mode, set the
    //    dead-band delay by loading ECCP1DEL<6:0>
    //    with the appropriate value.
    //   
    //    Bits PDC6:PDC0 of the ECCP1DEL register (Register 16-2) set the delay
    //    period in terms of microcontroller instruction cycles (TCY or 4 TOSC).
    
    ECCP1DEL = (ECCP1DEL & 0b10000000);
    
  
    // 9. Configure and start TMR2:
    //    - Clear the TMR2 interrupt flag bit by clearing
    //      the TMR2IF bit (PIR1<1>).
    //    - Set the TMR2 prescale value by loading the
    //      T2CKPS bits (T2CON<1:0>).
    //    - Enable Timer2 by setting the TMR2ON bit (T2CON<2>).
      
    // Timer2 for PWM
    OpenTimer2( 
        TIMER_INT_OFF & 
        T2_PS_1_16    // 1:1, 1:4, 1:16
      ); 
                
    // 10.Enable PWM outputs after a new PWM cycle
    //    has started:
    //    - Wait until TMRn overflows (TMRnIF bit is set).
      
      // ?
      
    //    - Enable the CCP1/P1A, P1B, P1C and/or P1D
    //      pin outputs by clearing the respective TRIS
    //      bits.
    
    TRISCbits.TRISC2 = 0; // P1A
    TRISDbits.TRISD5 = 0; // P1B
    
    //    - Clear the ECCPASE bit (ECCP1AS<7>).
    //      The modules output mode and polarity are
    //      configured by setting the P1M1:P1M0 and
    //      CCP1M3:CCP1M0 bits of the CCP1CON register.
  
    CCP1ASbits.ECCPASE = 0;
    
    protocol_set_cmd(PCMD_IDLE);
  }
  

void protocol_set_cmd(unsigned int CMD)
  {
    static unsigned int last_cmd = 0u;
    unsigned int new_dc = 0u;
    
    if (last_cmd == CMD) return;
    
    // cmd is a value between 0 <= CMD <= 255
    PR2 = CMD; 
    
    // adjust pulse width
    new_dc = 2*(CMD+1);
    SetDC( new_dc );
    
    last_cmd = CMD;
  }

#define WITHIN(X,LO,HI) ((X > LO) && (X < HI))

unsigned int protocol_get_cmd(void)
  {
    static unsigned int protocol_cmd = 0;
    static unsigned int cmd2 = 0;
    unsigned int cmd = 0;
    
    if (protocol_value)
      {
        // we have a command value
        if (WITHIN(protocol_value, PCMD_START_LO, PCMD_START_HI))
          {
            cmd = PCMD_START;
          }
        else if (WITHIN(protocol_value, PCMD_STOP_LO, PCMD_STOP_HI))
          {
            cmd = PCMD_STOP;
          }
        else if (WITHIN(protocol_value, PCMD_RESET_LO, PCMD_RESET_HI))
          {
            cmd = PCMD_RESET;
          }
        else if (WITHIN(protocol_value, PCMD_IDLE_LO, PCMD_IDLE_HI))
          {
            cmd = PCMD_IDLE;
          }          
        else
          {
            protocol_value = 0;
            return PCMD_UNKNOWN;
          }
        
        protocol_value = 0;
        
        // see if we need to update our current command        
        if (cmd == protocol_cmd)
          {
            // the command is the same as now. 
            cmd2 = cmd;
          }  
        else
          {
            // the command is different than the one we currently have
            if (cmd == cmd2)
              {
                // this makes two in a row!  Let's change to this command.
                protocol_cmd = cmd;
              }
            else
              {
                // this is not the same as the last one we tested.
                // but, we don't have two in a row.
                cmd2 = cmd;                
              }     
          }
      }
                          
    return protocol_cmd;
  }  


        
        // PWM Period   = [(PR2) + 1]  4  TOSC  (TMR2 Prescale Value)     (seconds/pwm_cycle)
        // INST_PER_SEC = Clock/4                                            (instruction/second)
        //
        // INST_PER_PWM_CYCLE = [PWM Period (sec/pwm cycle)] * (INST_PER_SEC instruction/sec)
        //                    = [(PR2) + 1]  4  TOSC  (TMR2 Prescale Value) * Clock/4
        //                    = [(PR2) + 1] * TOSC * (TMR2 Prescale Value) * Clock
        //                    = [(PR2) + 1] * (1/Clock) * (TMR2 Prescale Value) * Clock
        //                    = [(PR2) + 1] * (TMR2 Prescale Value)
        //                    = (new_pr2+1)*4;
        //
        // 

        /*
        if (VALUE_VSOURCE > VSOURCE_LIMIT_HIGH)
          {
            SetDeadBandDelay( (new_pr2+1)*2 );  
          }
        else if (VALUE_VSOURCE < VSOURCE_LIMIT_LOW)
          {
            SetDeadBandDelay( 0u );
          }          
          */     
        