

#include <timers.h>
#include "Compiler.h"
#include "GenericTypeDefs.h"
#include "HardwareProfile.h"
#include "usb_config.h"
#include "USB/usb_device.h"
#include "USB/usb.h"
#include "USB/usb_function_generic.h"

#include "user.h"
#include "user_common.h"
#include "sensors.h"
#include "my_sd_card.h"
#include "eeprom.h"
#include "protocol.h"

#include "eeprom_sector_value.h"
#include "toggle.h"


#pragma udata

/* states for the mail program functions */

#define ST_BEGIN      0
#define ST_START      1
#define ST_START_WAIT 2
#define ST_RUN        3
#define ST_WRITE      4
#define ST_STOP       5
#define ST_RESET      6
#define ST_RESET_WAIT 7
#define ST_USB        8

#define PTIMEOUT 5


volatile unsigned int tmr0_overflow_counter = 0;
volatile unsigned int tmr0_ptimeout = 0;

static BYTE protocol_test_mode = 0;

    
extern unsigned int protocol_value;    // from protocol.c    
extern BYTE sensor_update_complete; // from sensors.c
extern unsigned int SENSOR_VALUE;

TOGGLE_BUTTON start_stop_button;

#pragma udata USB_VARS
DATA_PACKET INPacket;
DATA_PACKET OUTPacket;
#pragma udata

USB_HANDLE USBGenericOutHandle = 0;
USB_HANDLE USBGenericInHandle = 0;



void ServiceRequests(void);


#pragma code

int SD_Card_Initialize(void)
  {
    int i;
    BYTE ret;
    
    ShutdownMedia();
    for (i=0; i<4; i++)
      {
        ret = MediaInitialize();
        
        if (ret) return 1;
      }
    return 0;    
  }

void SD_Card_Init(void)
  {
    LED_YELLOW_OFF; 
    while( !SD_Card_Initialize() ) LED_YELLOW_FLIP; 
    LED_YELLOW_OFF; 
  }
    

void UserInit(void)
  {    
    
    toggle_init(&start_stop_button, TOGGLE_STOP);
    
    sensors_init();    
    
    // http://www.camotruck.net/rollins/electronics/pic_timer.html
    
    // timer 0 serves as a "wall clock"
    OpenTimer0( 
        TIMER_INT_ON & 
        T0_SOURCE_INT &
        T0_16BIT &
        T0_PS_1_8  // 0.044 seconds until overflow
      );
            
    INTCON=0;    //make sure interrupts are disable
    INTCONbits.GIE=1;  //enable global interrupts
    INTCONbits.PEIE=1;  //enable peripheral interrupts
    INTCONbits.TMR0IE=1; //enable TMR0 overflow interrupt enable bit
    //INTCON=0xe0;   //all 4 lines above combined

    // initialize the protocol
    protocol_init();

    // initialize the SD card
    SD_Card_Init();
              
  }//end UserInit


/* return the number of t0 ticks since the last time this function was called */
static unsigned long read_timer0_delta(void)
  {
    static unsigned int overflow_previous = 0;
    unsigned int overflow_current = 0;
    
    static unsigned int time_previous = 0;    
    unsigned int time_current = 0;
    
    unsigned long value = 0;
        
    // read the timer
    // we first record the overflow counter value 
    overflow_current = tmr0_overflow_counter;
            
    // next we record the timer0 value
    time_current = ReadTimer0();
    
    // check whether the timer overflowed since we began this calculation
    if (overflow_current != tmr0_overflow_counter)
      {
        // we did overflow, so adjust the values
        overflow_current = tmr0_overflow_counter;
        time_current = 0;
      }  
      
    // since these are both unsigned, I should be able to do a simple "-" ?
    if (overflow_current > overflow_previous)
      {
        // this is the normal case.
        value = ((unsigned long)(time_current)) + (        ((unsigned long)overflow_current) -((unsigned long)overflow_previous))*65536L - ((unsigned long)(time_previous)); 
      }
    else
      {
        // the tmr0_overflow_counter overflowed.
        value = ((unsigned long)(time_current)) + ((65536L+((unsigned long)overflow_current))-((unsigned long)overflow_previous))*65536L - ((unsigned long)(time_previous)); 
      }
    
    time_previous = time_current;
    overflow_previous = overflow_current;
    return value;
  }

// these are used to make the state machine a little easier to follow

#define IF_P2(X,Y)         \
   if (X)                  \
    {                      \
      state=Y;             \
      break;               \
    }
    
#define IF_PCMD2(X,Y)     \
  if (PCMD == X)          \
    {                     \
      state=Y;            \
      break;              \
    }
    
#define IF_P3(X,Y,Z)       \
   if (X)                  \
    {                      \
      state=Y;             \
      protocol_set_cmd(Z); \
      break;               \
    }
    
#define IF_PCMD3(X,Y,Z)   \
  if (PCMD == X)          \
    {                     \
      state=Y;            \
      toggle_set_value(&start_stop_button,Z); \
      break;              \
    }
            
void ProcessIO(void)
  {       
    RECORD_TYPE rec; 
    static BYTE state = ST_BEGIN;
    static unsigned long sector_value = 0;
    unsigned int PCMD = 0;
    static unsigned int j = 0;
    
    BYTE P_START;
    BYTE P_STOP;
    
    // check the state of the start_stop toggle switch
    toggle_update( &start_stop_button, R_START_STOP);    

    // update the P_START and P_STOP flags    
    if (toggle_get_value( &start_stop_button ) == TOGGLE_STOP)
      {
        P_START = 0;
        P_STOP  = 1;
      }
    else
      {
        P_START = 1;
        P_STOP  = 0;
      }
  
    // get the comand from the protocol function
    if (!protocol_test_mode)
        PCMD = protocol_get_cmd();
        
    switch (state)
      {
        case ST_BEGIN:

            protocol_test_mode = 0;

            IF_P3(        P_USB,   ST_USB,    PCMD_STOP)    
                               
            IF_P3(        P_RESET, ST_RESET,  PCMD_RESET)            
            IF_PCMD2(  PCMD_RESET, ST_RESET) 
                       
            IF_P3(        P_START, ST_START,  PCMD_START)                      
            IF_PCMD3(  PCMD_START, ST_START,  TOGGLE_START)
            
            IF_P3(        P_STOP,  ST_BEGIN,  PCMD_STOP)  
            IF_PCMD3(  PCMD_STOP,  ST_BEGIN,  TOGGLE_STOP)
            
          break;
                  
        case ST_USB:
            
            IF_P3( !P_USB, ST_BEGIN, TOGGLE_STOP)
            
          break;   
          
        case ST_START: 
        
            tmr0_ptimeout = PTIMEOUT;
            
            LED_RED=LED_ON;
            
            sensors_start();

            state=ST_START_WAIT;
            
          break;
          
        case ST_START_WAIT:
            // when transistioning from BEGIN to RUN, the PCMD_STOP could be a little slow to update.
            // this state ensures it does update before going to the run state. 
            if (tmr0_ptimeout) break;
            
            // move sector point to last saved location
            EEPROM_read_sector_number( &sector_value );
            SD_set_sector_value(sector_value);
            stream_init(sector_value);
            
            read_timer0_delta(); // perform a read to reset the delta counter
                        
            state=ST_RUN;
          break;
                      
        case ST_RUN:

            IF_P3(        P_STOP,  ST_STOP,  PCMD_STOP)            
            IF_PCMD3(  PCMD_STOP,  ST_STOP,  TOGGLE_STOP) 
                       
            IF_P3(        P_USB,   ST_STOP,  PCMD_STOP) 
            
            IF_P3(        P_RESET, ST_RESET, PCMD_RESET)                 
            IF_PCMD2(  PCMD_RESET, ST_RESET)
         
            sensors_update();
            
            stream_update();

            if (sensor_update_complete) 
              {
                state=ST_WRITE;
                sensor_update_complete = 0;
              }
          break;
          
        case ST_WRITE:
        
            rec.delay = (unsigned int)(read_timer0_delta()/4L);
            
            // read the sensor value
            //rec.value = SENSOR_VALUE;
               rec.value = j++;         
            // if ( ! SD_write(&rec) ) Nop(); //LED_RED_FLIP;  // write error
            stream_write( (BYTE*) &rec, 4);
            
            state=ST_RUN;

          break;
           
        case ST_STOP:
            toggle_set_value(&start_stop_button,TOGGLE_STOP);
            
            // flush the sector buffer
            // zeros will pad the sector
            // if ( ! SD_flush() ) Nop();// LED_RED_FLIP;
            stream_close();
            
            // write sector number to eeprom
           // sector_value = SD_get_sector_value();
            
            sector_value = get_stream_sector_number();
            EEPROM_write_sector_number( &sector_value );
            
            LED_RED=LED_OFF;            
            state=ST_BEGIN;
          break;
          
        case ST_RESET:
            toggle_set_value(&start_stop_button,TOGGLE_STOP);
            
            LED_RED=LED_OFF;
            LED_YELLOW=LED_ON;    

            tmr0_ptimeout = PTIMEOUT;

            stream_close();
                            
            SD_erase(); 
            sector_value = 0;
            
            EEPROM_write_sector_number( &sector_value );            
            sensors_start();
            
            SD_Card_Init();
            state = ST_RESET_WAIT;
          break;
        
        case ST_RESET_WAIT:
            // NOTE: wait until the reset button is released
            if (P_RESET) break;
            
            // wait for the ptotocol timeout to expire
            if (tmr0_ptimeout) break;
            
            protocol_set_cmd(PCMD_IDLE);
            
            if (PCMD == PCMD_RESET) break;
            
            LED_YELLOW=LED_OFF;             
            
            state = ST_BEGIN;
          break;
      } // switch
      
    if((USBDeviceState < CONFIGURED_STATE)||(USBSuspendControl==1)) return;
    
    ServiceRequests();

}//end ProcessIO


void ServiceRequests(void)
  {
    static BYTE sdtest = 0;  
    unsigned int p_cmd;  
    unsigned int command;
    BYTE i;
    BYTE n;
    BYTE ret;  
    RECORD_TYPE rec; 
    RECORD_TYPE *ptr;
    DWORD CardCapacity;
    DWORD SectorSize;
    static unsigned long sector_count = 0;

    //Check to see if data has arrived
    if(!USBHandleBusy(USBGenericOutHandle))
      {        
        command = OUTPacket.b[0]; 

        switch (command) 
          {            
            case CMD_REQ_PING:
                pack_C(&INPacket, CMD_ANS_PONG);
                
                if(!USBHandleBusy(USBGenericInHandle))
                  USBGenericInHandle = USBGenWrite(USBGEN_EP_NUM,(BYTE*)&INPacket,DATA_PACKET_SIZE);  
                break;        
                                
            case CMD_REQ_SDTEST_PUT: 
            
                if (sdtest != CMD_REQ_SDTEST_PUT) 
                  {
                    SD_erase();
                    sdtest = CMD_REQ_SDTEST_PUT;
                  }
                unpack_C22(&OUTPacket, (BYTE*) &command, (UINT16*) &rec.delay, (UINT16*) &rec.value);
                
                if ( SD_write(&rec))
                    pack_C(&INPacket,CMD_OK);
                else
                    pack_C(&INPacket,CMD_ERR);                
               
                if(!USBHandleBusy(USBGenericInHandle))
                  USBGenericInHandle = USBGenWrite(USBGEN_EP_NUM,(BYTE*)&INPacket,DATA_PACKET_SIZE);
                break;                        
                      
            case CMD_REQ_SDTEST_GET:

                if (sdtest != CMD_REQ_SDTEST_GET) 
                  {
                    SD_seek_sof();
                    sdtest = CMD_REQ_SDTEST_GET;
                  }
                                                        
                if (ptr = SD_read(10))
                  {
                    //ptr->delay=13; ptr->value=13;            
                    pack_C22(&INPacket, CMD_ANS_DATA, ptr->delay, ptr->value);  
                  }
                else
                  {                
                    pack_C(&INPacket, CMD_ERR_EOF);
                  } 

                if(!USBHandleBusy(USBGenericInHandle))
                  USBGenericInHandle = USBGenWrite(USBGEN_EP_NUM,(BYTE*)&INPacket,DATA_PACKET_SIZE);
                break;                        

            case CMD_REQ_SEC_COUNT:
            
                // read the value for the sector from EEPROM      
                EEPROM_read_sector_number( &sector_count );    
                           
                pack_C4(&INPacket, CMD_ANS_DATA, (UINT32) sector_count);

                if(!USBHandleBusy(USBGenericInHandle))
                  USBGenericInHandle = USBGenWrite(USBGEN_EP_NUM,(BYTE*)&INPacket,DATA_PACKET_SIZE);
                break;

            case CMD_REQ_SIZE:
                        
                CardCapacity = ReadCapacity();
                SectorSize   = ReadSectorSize();
                      
                pack_C44(&INPacket, CMD_ANS_DATA, (UINT32) CardCapacity, (UINT32) SectorSize);
                                                           

                if(!USBHandleBusy(USBGenericInHandle))
                  USBGenericInHandle = USBGenWrite(USBGEN_EP_NUM,(BYTE*)&INPacket,DATA_PACKET_SIZE);
                break;
                                
            case CMD_REQ_ERASE:
                LED_YELLOW = LED_ON;
                // 2097152
                SD_CUSTOM_Sector_Erase(0, SECTOR_COUNT_MAX);
                LED_YELLOW = LED_OFF;
                
                pack_C(&INPacket,CMD_OK);

                if(!USBHandleBusy(USBGenericInHandle))
                  USBGenericInHandle = USBGenWrite(USBGEN_EP_NUM,(BYTE*)&INPacket,DATA_PACKET_SIZE);
                
                break;

            case CMD_REQ_SOF:
            
                SD_Card_Init();
                
                // move to the first of the sd card
                SD_seek_sof();                
                
                // read the total number of sectors
                EEPROM_read_sector_number( &sector_count );
                
                pack_C(&INPacket,CMD_OK);

                if(!USBHandleBusy(USBGenericInHandle))
                  USBGenericInHandle = USBGenWrite(USBGEN_EP_NUM,(BYTE*)&INPacket,DATA_PACKET_SIZE);
                break;                                    
            
            case CMD_REQ_BULK:
                                                      
                if (ptr = SD_read(sector_count))
                  {
                    // add the command to the bulk packet
                    pack_bulk_C(&INPacket, CMD_ANS_BULK);  
                       
                    // now add 15 records 
                    // note that the first record has already been read.                             
                    i=0;
                    while (ptr && (i<14))
                      {                                                
                        pack_bulk_n22(&INPacket, i, ptr->delay, ptr->value); //i=[0..14]
                        ptr = SD_read(sector_count);
                        i++;
                      }     
                    // i = 14: we need to insert the last record.
                    if (ptr) pack_bulk_n22(&INPacket, 14, ptr->delay, ptr->value); //i=[0..14]               
                  }
                else
                  {                
                    pack_bulk_C(&INPacket, CMD_ERR_EOF);
                  } 

                if(!USBHandleBusy(USBGenericInHandle))
                  USBGenericInHandle = USBGenWrite(USBGEN_EP_NUM,(BYTE*)&INPacket,DATA_PACKET_SIZE);
                break;                        
                       
            case CMD_SET_PTEST:
            
                unpack_C2(&OUTPacket, (BYTE*) &command, (UINT16*) &p_cmd);
           
                if (p_cmd < 255)
                  {
                    protocol_set_cmd(p_cmd);
                    protocol_test_mode = 1;                    
                    pack_C(&INPacket,CMD_OK);
                  }
                else
                    pack_C(&INPacket,CMD_ERR);                
               
                if(!USBHandleBusy(USBGenericInHandle))
                  USBGenericInHandle = USBGenWrite(USBGEN_EP_NUM,(BYTE*)&INPacket,DATA_PACKET_SIZE);
                break;                        

            case CMD_GET_PTEST:
           
                pack_C2(&INPacket,CMD_VAL_PTEST,protocol_value);                
               
                if(!USBHandleBusy(USBGenericInHandle))
                  USBGenericInHandle = USBGenWrite(USBGEN_EP_NUM,(BYTE*)&INPacket,DATA_PACKET_SIZE);
                break;      

                
          } // switch
          
        //Re-arm the OUT endpoint for the next packet
        USBGenericOutHandle = USBGenRead(USBGEN_EP_NUM,(BYTE*)&OUTPacket,USBGEN_EP_SIZE);
      }//end if
  }//end ServiceRequests




